import datetime as dt
import uuid
import logging
import json

from sqlalchemy import Column, Unicode, DateTime
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.dialects.postgresql import JSON

from ckan.model.meta import metadata

log = logging.getLogger(__name__)


def make_uuid():
    return str(uuid.uuid4())


Base = declarative_base(metadata=metadata)


class Datapackage(Base):
    __tablename__ = u'datapackage'

    id = Column(Unicode, primary_key=True, default=make_uuid)
    package_id = Column(Unicode)
    status = Column(Unicode, default=u'created')
    created = Column(DateTime, default=dt.datetime.utcnow)
    data = Column(JSON)
    errors = Column(JSON)

    def errors_json(self):
        return json.dumps(self.errors)


class DatapackageResource(Base):
    __tablename__ = u'datapackage_resource'

    id = Column(Unicode, primary_key=True, default=make_uuid)
    resource_id = Column(Unicode)
    status = Column(Unicode, default=u'created')
    created = Column(DateTime, default=dt.datetime.utcnow)
    data = Column(JSON)
    errors = Column(JSON)


def create_tables():
    Datapackage.__table__.create()
    DatapackageResource.__table__.create()



import string

from functools import reduce

from ckan.logic import ValidationError

from ckanext.datapackage_creator.settings import settings


def validate_resource(data):
    errors = {}
    name = data.get('name', '')
    check_slug = lambda x, y: x and (y in string.ascii_lowercase or y in ['.', '-', '_'] or y.isdigit())
    if not name:
        errors['Name'] = ['This field is required']
    elif not reduce(check_slug, name, True):
        errors['Name'] = [
            'It must consist only of lowercase alphanumeric characters plus “.”, “-” and “_”.'
        ]
    resource_required = settings.get('resource', {}).get('required', [])
    resource_fields_required = [field for field in resource_required if not field.startswith('field.')]
    fields_required = [field for field in resource_required if field.startswith('field.')]
    for field in resource_fields_required:
        value = data.get(field)
        if not value:
            errors[field.capitalize()] = ['This field is required']
    for i, field in enumerate(data.get('fields', [])):
        index = i + 1
        errors_field = []
        for field_required in fields_required:
            field_required = field_required.replace('field.', '')
            value = field.get(field_required)
            if not value:
                errors_field.append(f'{field_required.capitalize()} is required')
        if errors_field:
            errors[f'Table Schema {index}'] = [', '.join(errors_field)]
    if errors:
        error_summary = {k: ', '.join(v) for k, v in errors.items()}
        raise ValidationError(errors=errors, error_summary=error_summary)


def validate_package(data):
    errors = {}
    title = data.get('title')
    if not title:
        errors['Title'] = ['This field is required']
    owner_org = data.get('owner_org')
    if not owner_org:
        errors['Organization'] = ['This field is required']
    package_required = settings.get('package', {}).get('required', [])
    map_fields = {
        'license': 'license_id',
        'source': 'url',
        'description': 'notes',
    }
    for field in package_required:
        real_field = map_fields.get(field, field)
        value = data.get(real_field)
        if not value:
            errors[field.capitalize().replace('_', ' ')] = ['This field is required']
    if errors:
        error_summary = {k: ', '.join(v) for k, v in errors.items()}
        raise ValidationError(errors=errors, error_summary=error_summary)



import frictionless

import ckan.model as model
from ckan.common import config

from ckanext.datapackage_creator.model import Datapackage


class BaseBackend(object):

    def describe_resource(self, resource, *args, **kwargs):
        raise NotImplementedError()

    def extract_resource(self, resource, *args, **kwargs):
        raise NotImplementedError()

    def validate_package(self, package, *args, **kwargs):
        raise NotImplementedError()


class FritionlesseBackend(BaseBackend):

    def describe_resource(self, resource, *args, **kwargs):
        return frictionless.describe_resource(resource, *args, **kwargs)

    def extract_resource(self, resource, *args, **kwargs):
        return frictionless.extract_resource(resource, *args, **kwargs)

    def validate_package(self, package, *args, **kwargs):
        return frictionless.validate(package)


default = FritionlesseBackend()



import os
import json
import tempfile

from flask import make_response, request

from slugify import slugify

import ckan.model as model
import ckan.plugins.toolkit as toolkit

from ckan.logic import get_action, ValidationError
from ckan.views.dataset import _tag_string_to_list

from ckanext.datapackage_creator.utils import row_to_dict
from ckanext.datapackage_creator.model import Datapackage, DatapackageResource
from ckanext.datapackage_creator.validation import validate_resource, validate_package
from ckanext.datapackage_creator.settings import settings
from ckanext.datapackage_creator.backends import default as default_backend


def inference():
    context = {
        'model': model,
        'session': model.Session,
        'user': toolkit.c.user,
        'auth_user_obj': toolkit.c.userobj,
    }
    try:
        toolkit.check_access('package_create', context)
    except toolkit.NotAuthorized:
        toolkit.abort(401, toolkit._('Unauthorized to create a dataset'))
    response = make_response()
    response.content_type = 'application/json'
    file = request.files['file']
    _, extension = os.path.splitext(file.filename)
    tmp = tempfile.NamedTemporaryFile(suffix=extension, delete=False)
    file.save(tmp)
    tmp.close()
    action = get_action('inference_data')
    data = {
        'filepath': tmp.name,
    }
    try:
        result = action(context, data)
        result['has_error'] = False
        result['error_summary'] = ''
    except:
        result = {
            'has_error': True,
            'error_summary': 'Error in data inference'
        }
    else:
        result['has_error'] = False
        result['error_summary'] = ''
        name, ext = os.path.splitext(file.filename)
        result['metadata']['name'] = slugify(name)
    response.data = json.dumps(result)
    return response


def settings_show():
    context = {
        'model': model,
        'session': model.Session,
        'user': toolkit.c.user,
        'auth_user_obj': toolkit.c.userobj,
        'api_version': 3,
        'for_edit': True,
    }
    try:
        toolkit.check_access('package_create', context)
    except toolkit.NotAuthorized:
        toolkit.abort(401, toolkit._('Unauthorized to create a dataset'))
    response = make_response()
    response.content_type = 'application/json'
    options = {
        'resource': settings.get('resource', {'required': [], 'editable': []}),
        'package': settings.get('package', {'required': [], 'editable': []}),
    }
    response.data = json.dumps(options)
    return response


def save_resource():
    context = {
        'model': model,
        'session': model.Session,
        'user': toolkit.c.user,
        'auth_user_obj': toolkit.c.userobj,
        'api_version': 3,
        'for_edit': True,
    }
    try:
        toolkit.check_access('package_create', context)
    except toolkit.NotAuthorized:
        toolkit.abort(401, toolkit._('Unauthorized to create a dataset'))
    data = request.form.copy()
    data['state'] = 'active'
    data['url_type'] = 'upload'
    resource_id = data.get('id')
    if not resource_id:
        data['upload'] = request.files['upload']
    metadata = data['metadata']
    del data['metadata']
    data_response = {
        'has_error': False,
        'resource': None
    }
    data_validate = data.copy()
    data_validate['fields'] = json.loads(metadata).get('fields')
    try:
        validate_resource(data_validate)
        if resource_id:
            action = get_action('resource_patch')
        else:
            action = get_action('resource_create')
        resource = action(context, data)
    except ValidationError as e:
        data_response['errors'] = e.error_dict
        data_response['error_summary'] = e.error_summary
        data_response['has_error'] = True
    else:
        data_response['resource'] = resource
        save_datapackage_resource = get_action('save_datapackage_resource')
        data_package_resource = {
            'metadata': metadata,
            'resource_id': resource['id'],
            'errors': {}
        }
        save_datapackage_resource(context, data_package_resource)
    response = make_response()
    response.content_type = 'application/json'
    response.data = json.dumps(data_response)
    return response


def delete_resource(resource_id):
    context = {
        'model': model,
        'session': model.Session,
        'user': toolkit.c.user,
        'auth_user_obj': toolkit.c.userobj,
        'api_version': 3,
        'for_edit': True,
    }
    try:
        toolkit.check_access('package_create', context)
    except toolkit.NotAuthorized:
        toolkit.abort(401, toolkit._('Unauthorized to create a dataset'))
    data = {
        'id': resource_id,
        'state': 'deleted'
    }
    data_response = {
        'has_error': False,
        'package': None
    }
    try:
        resource = get_action('resource_patch')(context, data)
    except ValidationError as e:
        data_response['errors'] = e.error_dict
        data_response['error_summary'] = e.error_summary
        data_response['has_error'] = True
    response = make_response()
    response.content_type = 'application/json'
    response.data = json.dumps(data_response)
    return response


def save_package():
    context = {
        'model': model,
        'session': model.Session,
        'user': toolkit.c.user,
        'auth_user_obj': toolkit.c.userobj,
        'api_version': 3,
        'for_edit': True,
    }
    try:
        toolkit.check_access('package_create', context)
    except toolkit.NotAuthorized:
        toolkit.abort(401, toolkit._('Unauthorized to create a dataset'))
    data = request.form.copy()
    tag_string = data.pop('tag_string')
    data['tags'] = _tag_string_to_list(tag_string)
    package_id = data.get('id')
    if not package_id:
        data['state'] = 'draft'
        data['_ckan_phase'] = 'dataset_new_1'
        package_action = get_action('package_create')
    else:
        package_action = get_action('package_patch')
    metadata = data.pop('metadata')
    metadata_json = json.loads(metadata)
    data['extras'] = metadata_json['extras']
    data_response = {
        'has_error': False,
        'package': None,
    }
    try:
        validate_package(data)
        package = package_action(context, data)
    except ValidationError as e:
        data_response['errors'] = e.error_dict
        data_response['error_summary'] = e.error_summary
        data_response['has_error'] = True
    else:
        data_response['package'] = package
        save_datapackage = get_action('save_datapackage')
        data_datapackage = {
            'package_id': package['id'],
            'metadata': metadata
        }
        save_datapackage(context, data_datapackage)
    response = make_response()
    response.content_type = 'application/json'
    response.data = json.dumps(data_response)
    return response


# def validate_package():
#     context = {
#         'model': model,
#         'session': model.Session,
#         'user': toolkit.c.user,
#         'auth_user_obj': toolkit.c.userobj,
#         'api_version': 3,
#         'for_edit': True,
#     }
#     try:
#         toolkit.check_access('package_create', context)
#     except toolkit.NotAuthorized:
#         toolkit.abort(401, toolkit._('Unauthorized to create a dataset'))


def publish_package():
    context = {
        'model': model,
        'session': model.Session,
        'user': toolkit.c.user,
        'auth_user_obj': toolkit.c.userobj,
        'api_version': 3,
        'for_edit': True,
    }
    try:
        toolkit.check_access('package_create', context)
    except toolkit.NotAuthorized:
        toolkit.abort(401, toolkit._('Unauthorized to create a dataset'))
    data_response = {
        'has_error': False,
    }
    data = request.form.copy()
    data['state'] = 'active'
    package_data = {
        'id': data['id'],
        'state': 'active'
    }
    metadata = data['metadata']
    try:
        get_action('package_patch')(context, package_data)
        frictionless_package = get_action('generate_datapackage_json')(context, package_data)
    except ValidationError as e:
        data_response['errors'] = e.error_dict
        data_response['error_summary'] = e.error_summary
        data_response['has_error'] = True
    else:
        validation = default_backend.validate_package(frictionless_package)
        datapackage = model.Session.query(Datapackage).filter(
            Datapackage.package_id==package_data['id']
        ).order_by(Datapackage.created.desc()).first()
        new_datapackage = Datapackage()
        new_datapackage.package_id = package_data['id']
        new_datapackage.errors = validation.to_dict()
        if datapackage:
            new_datapackage.data = datapackage.data
        else:
            new_datapackage.data = json.loads(metadata)
        model.Session.add(new_datapackage)
        model.Session.commit()
    response = make_response()
    response.content_type = 'application/json'
    response.data = json.dumps(data_response)
    return response


def datapackage_show(package_id):
    context = {
        'model': model,
        'session': model.Session,
        'user': toolkit.c.user,
        'auth_user_obj': toolkit.c.userobj,
        'api_version': 3,
    }
    try:
        toolkit.check_access('package_show', context)
    except toolkit.NotAuthorized:
        toolkit.abort(401, toolkit._('Unauthorized to create a dataset'))
    response = make_response()
    response.content_type = 'application/json'
    data = {
        'package_id': package_id
    }
    try:
        datapackage = get_action('datapackage_show')(context, data)
    except Exception as ex:
        datapackage = Datapackage()
    data = {
        'id': package_id
    }
    package = get_action('package_show')(context, data)
    datapackage_json = row_to_dict(datapackage)
    datapackage_json['errors_json'] = datapackage.errors_json()
    data_response = {
        'datapackage': datapackage_json,
        'package': package,
    }
    response.data = json.dumps(data_response)
    return response


def datapackage_resource_show(resource_id):
    context = {
        'model': model,
        'session': model.Session,
        'user': toolkit.c.user,
        'auth_user_obj': toolkit.c.userobj,
        'api_version': 3,
    }
    try:
        toolkit.check_access('package_show', context)
    except toolkit.NotAuthorized:
        toolkit.abort(401, toolkit._('Unauthorized to create a dataset'))
    response = make_response()
    response.content_type = 'application/json'
    data = {
        'resource_id': resource_id
    }
    try:
        datapackage_resource = get_action('datapackage_resource_show')(context, data)
    except:
        datapackage_resource = DatapackageResource()
    data = {
        'id': resource_id
    }
    resource = get_action('resource_show')(context, data)
    data_response = {
        'datapackage_resource': row_to_dict(datapackage_resource),
        'resource': resource,
    }
    response.data = json.dumps(data_response)
    return response


def datapackage_json_show(package_id):
    context = {
        'model': model,
        'session': model.Session,
        'user': toolkit.c.user,
        'auth_user_obj': toolkit.c.userobj,
        'api_version': 3,
    }
    data = {
        'id': package_id
    }
    frictionless_package = get_action('generate_datapackage_json')(context, data)
    response = make_response()
    response.content_type = 'application/json'
    response.data = json.dumps(frictionless_package)
    return response


def datapackage_validation_show(package_id):
    context = {
        'model': model,
        'session': model.Session,
        'user': toolkit.c.user,
        'auth_user_obj': toolkit.c.userobj,
        'api_version': 3,
    }
    data = {
        'id': package_id
    }
    package = get_action('package_show')(context, data)
    datapackage_list = model.Session.query(Datapackage).filter(
        Datapackage.package_id==package['id']
    ).order_by(Datapackage.created.desc())
    return toolkit.render(
        'datapackage_creator/validation_read.html',
        extra_vars={
            'datapackage_list': datapackage_list,
            'pkg_dict': package
        }
    )



import json

from ckan.model import Session

from frictionless_ckan_mapper import ckan_to_frictionless as converter

from ckanext.datapackage_creator.model import Datapackage, DatapackageResource


def ckan_resource_to_frictionless(resource):
    frictionless_resource = converter.resource(resource)
    frictionless_resource['path'] = resource['url']
    datapackage_resource = Session.query(DatapackageResource).filter(
        DatapackageResource.resource_id==resource['id']
    ).order_by(DatapackageResource.created.desc()).first()
    if datapackage_resource:
        extras = json.loads(datapackage_resource.data)
        frictionless_resource['schema'] = {
            'fields': [],
        }
        foreign_keys = []
        try:
            fields = extras['inference']['metadata']['schema']['fields']
        except:
            pass
        for field in fields:
            field_dict = {
                'name': field.get('name', ''),
                'description': field.get('description', ''),
                'title': field.get('title', ''),
                'format': field.get('format'),
                'description': field.get('description', ''),
                'type': field.get('type', ''),
            }
            primary_key = field.get('primary_key')
            if primary_key:
                frictionless_resource['schema']['primary_key'] = field['name']
            foreign_key = field.get('foreign_key')
            if foreign_key:
                try:
                    resource, resource_field = foreign_key.split()
                except:
                    pass
                else:
                    foreign_keys.append(
                        {
                            'fields': field['name'],
                            'reference': {
                                "resource": resource,
                                "fields": resource_field
                            }
                        }
                    )
            frictionless_resource['schema']['fields'].append(field_dict)
        if foreign_keys:
            frictionless_resource['foreignKeys'] = []
    return frictionless_resource


def ckan_to_frictionless(package):
    package_id = package['id']
    datapackage = Session.query(Datapackage).filter(
        Datapackage.package_id==package_id
    ).order_by(Datapackage.created.desc()).first()
    frictionless_package = converter.dataset(package)
    try:
        if datapackage:
            extras = json.loads(datapackage.data)
            frictionless_package['contributors'] = []
            for contributor in extras['contributors']:
                contributor_item = {
                    'title': contributor['name'],
                    'role': contributor['type'].lower(),
                    'email': contributor['email'],
                }
                if contributor['url'].strip():
                    contributor_item['path'] = contributor['url']
                frictionless_package['contributors'].append(contributor_item)
            frictionless_package['frequency'] = extras['frequency']
            for extra in package['extras']:
                frictionless_package[extra['key']] = extra['value']
        resources = []
        for resource in package['resources']:
            frictionless_resource = ckan_resource_to_frictionless(resource)
            resources.append(frictionless_resource)
        frictionless_package['resources'] = resources
    except:
        pass
    return frictionless_package
